#include "specimen.h"
#include "driver.h"
#include "adsr.h"
#include "maths.h"
#include "ticks.h"

void adsr_init(adsr_t *env)
{
     env->state = ADSR_STATE_IDLE;
     env->ticks   = 0;
     env->val     = 0.0;
     env->rval    = 0.0;
     env->attack  = 0;
     env->decay   = 0;
     env->sustain = 1.0;
     env->release = 0;
}

void adsr_trigger(adsr_t *env)
{
     env->state = ADSR_STATE_ATTACK;
     env->ticks = 0;
}

void adsr_release(adsr_t *env)
{
     env->state = ADSR_STATE_RELEASE;
     env->ticks = 0;
}

/* return current value between 0 and 1 and advance */
float adsr_tick(adsr_t *e)
{
     float d;

     switch(e->state) {
     case ADSR_STATE_ATTACK:
	  if (e->attack == 0) {
	       e->val = 1.0;
	  } else {
	       d = (e->ticks * 1.0) / e->attack;
	       e->val = lerp(0.0, 1.0, d);
	  }

	  if (++e->ticks > e->attack) {
	       e->state = ADSR_STATE_DECAY;
	       e->ticks = 0;
	  }
	  break;
     case ADSR_STATE_DECAY:
	  if (e->decay == 0) {
	       e->val = e->sustain;
	  } else {
	       d = (e->ticks * 1.0) / e->decay;
	       e->val = lerp(1.0, e->sustain, d);
	  }

	  if (++e->ticks > e->decay) {
	       e->state = ADSR_STATE_SUSTAIN;
	       e->ticks = 0;
	  }
	  break;
     case ADSR_STATE_SUSTAIN:
	  /* e->val just hovers at e->sustain */
	  break;
     case ADSR_STATE_RELEASE:
	  if (e->release == 0) {
	       e->val = 0.0;
	  } else if (e->ticks == 0) {
	       /* we are just beginning to release, store our current
		* value for later use */
	       e->rval = e->val;
	  } else {
	       d = (e->ticks * 1.0) / e->release;
	       e->val = lerp(e->rval, 0.0, d);
	  }

	  if (++e->ticks > e->release) {
	       e->state = ADSR_STATE_IDLE;
	       e->ticks = 0;
	  }
	  break;
     case ADSR_STATE_IDLE:		/* fall through */
     default:
	  e->val = 0.0;
	  break;
     }

     return e->val;
}

void  adsr_set_params (adsr_t *env, adsr_params_t *params)
{
     env->attack  = sec2tick(params->attack);
     env->decay   = sec2tick(params->decay);
     env->sustain = params->sustain;
     env->release = sec2tick(params->release);
}
